import sys
import os
from unittest.mock import MagicMock as Mock
from setuptools_scm import get_version

# Mocking to make RTD autobuild the documentation. (Doesn’t suffice.)
# autodoc_mock_imports = [ 'numpy', 'symengine', '.' ]

MOCK_MODULES = [
	'numpy', 'numpy.testing', 'numpy.random',
	'symengine', 'symengine.printing',
	'.'
]
sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

sys.path.insert(0,os.path.abspath("../jitcxde_common"))

needs_sphinx = '1.3'

extensions = [
		'sphinx.ext.autodoc',
		'sphinx.ext.autosummary',
		'sphinx.ext.mathjax',
		'numpydoc',
	]

source_suffix = '.rst'

master_doc = 'index'

project = u'JiTC*DE Common'
copyright = u'2017, Gerrit Ansmann'

release = version = get_version(root='..', relative_to=__file__)

default_role = "any"

add_function_parentheses = True

add_module_names = False

html_theme = 'nature'
pygments_style = 'colorful'
htmlhelp_basename = 'JiTC*DEdoc'

numpydoc_show_class_members = False
autodoc_member_order = 'bysource'

def on_missing_reference(app, env, node, contnode):
	if node['reftype'] == 'any':
		return contnode
	else:
		return None

def setup(app):
	app.connect('missing-reference', on_missing_reference)



from symengine.lib.symengine_wrapper import FunctionSymbol

def is_call(expression,function):
	"""
	whether expression is a call of function
	"""
	return (
			expression.__class__ == FunctionSymbol and
			expression.get_name() == function.name
		)

def function_visitor(expression,function):
	"""
	Generator function that yields all subexpressions of `expression` that are an instance of `function`.
	"""
	
	if is_call(expression,function):
		yield expression
	
	for arg in expression.args:
		yield from function_visitor(arg,function)

def collect_arguments(expression,function):
	"""
	Parameters
	----------
	expression : SymEngine expression
	function : SymEngine function
	
	Returns
	-------
	arguments : set of SymEngine expressions
		set of all arguments with which `function` is called within `expression`.
	"""
	
	return {
			expr.args
			for expr in function_visitor(expression,function)
		}

def count_calls(expression,function):
	"""
	Counts the calls of `function` within `expression`.
	Similar to SymPy’s `count`.
	"""
	return sum(1 for _ in function_visitor(expression,function))

def has_function(expression,function):
	"""
	Checks whether `function` is called within `expression`.
	Similar to SymPy’s `has`.
	"""
	
	try:
		next(function_visitor(expression,function))
	except StopIteration:
		return False
	else:
		return True

def ordered_subs(expression,substitutions):
	for substitution in substitutions:
		expression = expression.subs(*substitution)
	return expression

def replace_function(expression,function,new_function):
	"""
		Replaces all instances of `function` within `expression` with `new_function`.
		To bypass SymEngine issue #1356.
	"""
	if expression.is_Atom:
		return expression
	else:
		replaced_args = (
				replace_function(arg,function,new_function)
				for arg in expression.args
			)
		if is_call(expression,function):
			return new_function(*replaced_args)
		else:
			return expression.func(*replaced_args)




from itertools import chain
from symengine.printing import ccode
from jitcxde_common.strings import count_up

def codelines(expressions):
	"""
	Tries to convert expressions to working code
	"""
	for expression in expressions:
		try:
			codeline = ccode(expression)
		except RuntimeError as error:
			if "Not supported" not in str(error):
				raise
			else:
				raise NotImplementedError(
						"Cannot convert the following expression to C Code:\n"
						+ str(expression)
					)
		yield codeline + ";\n"

def render_declarator(name, _type, size=0):
	return _type + " " + name + ("[%i]"%size if size else "")

def write_in_chunks(lines,mainfile,deffile,name,chunk_size,arguments,omp=True):
	"""
		Writes lines to files in chunks if appropriate.
		
		Parameters
		----------
		lines: iterator
			the lines of code to be written
		mainfile: file object
			the file that contains the central calls
		deffile: file object
			the file where definitions of functions are included
		name: string
			unique name of what is computed
		chunk_size: integer
			size of chunks. If the number of lines is below this, they will plainly written to mainfile.
		arguments: list of tuples
			Each tuple contains the name, type, and size (optional, for arrays) of an argument needed by the code.
			This is so the arguments can be passed to chunked functions.
		omp: boolean
			whether OMP pragmas should be included
	"""
	funcname = "definitions_" + name
	
	first_chunk = []
	try:
		for _ in range(chunk_size+1):
			first_chunk.append(next(lines))
	except StopIteration:
		# No chunking:
		for line in first_chunk:
			mainfile.write(line)
	else:
		lines = chain(first_chunk, lines)
		
		if omp:
			mainfile.write("#pragma omp parallel sections\n{\n")
		
		while True:
			try:
				next_line = next(lines)
			except StopIteration:
				break
			
			if omp:
				mainfile.write("#pragma omp section\n")
			mainfile.write("{" + funcname + "(")
			deffile.write("void " + funcname + "(")
			if arguments:
				mainfile.write(", ".join(argument[0] for argument in arguments))
				deffile.write(", ".join(render_declarator(*argument) for argument in arguments))
			else:
				deffile.write("void")
			mainfile.write(");}\n")
			deffile.write("){\n")
			
			deffile.write(next_line)
			try:
				for _ in range(chunk_size-1):
					deffile.write(next(lines))
			except StopIteration:
				break
			finally:
				deffile.write("}\n")
			
			funcname = count_up(funcname)
		
		if omp:
			mainfile.write("}")




import unittest
from symengine import Function, Symbol, sin, Integer
from jitcxde_common.symbolic import collect_arguments, count_calls, has_function, replace_function

f = Function("f")
g = Function("g")
a = Symbol("a")

class TestCollectArguments(unittest.TestCase):
	def test_complex_expression(self):
		expression = 3**f(42) + 23 - f(43,44) + f(45+a)*sin( g(f(46,47,48)+17) - g(4) )
		
		self.assertEqual(
				collect_arguments(expression,f),
				{
					(Integer(42),),
					(Integer(43), Integer(44)),
					(45+a,),
					(Integer(46), Integer(47), Integer(48))
				}
			)
		
		self.assertEqual(count_calls(expression,f),4)
		self.assertTrue(has_function(expression,f))
		self.assertEqual(
				replace_function(expression,f,g),
				3**g(42) + 23 - g(43,44) + g(45+a)*sin( g(g(46,47,48)+17) - g(4) )
			)
	
	def test_function_within_function(self):
		expression = f(f(42))
		
		self.assertEqual(
				collect_arguments(expression, f),
				{ (Integer(42),) , (f(Integer(42)),) }
			)
		
		self.assertEqual(count_calls(expression,f),2)
		self.assertTrue(has_function(expression,f))
		self.assertEqual( replace_function(expression,f,g), g(g(42)) )
	
	def test_no_function(self):
		expression = g(a)+42
		
		self.assertEqual( collect_arguments(expression,f), set() )
		self.assertEqual(count_calls(expression,f),0)
		self.assertFalse(has_function(expression,f))
		self.assertEqual( replace_function(expression,f,g), g(a)+42 )


if __name__ == "__main__":
	unittest.main(buffer=True)



import numpy as np

def random_direction(n):
	"""
	returns an n-dimensional vector with random direction and length n
	"""
	vector = np.random.normal(0,1,n)
	return vector/np.linalg.norm(vector)

def orthonormalise(vectors):
	"""
	Orthonormalise vectors in place (with Gram-Schmidt) and return their norms after orthogonalisation (but before normalisation).
	"""
	norms = []
	for i,vector in enumerate(vectors):
		for j in range(i):
			vector -= np.dot( vector, vectors[j] ) * vectors[j]
		norm = np.linalg.norm(vector)
		vector /= norm
		norms.append(norm)
	
	return np.array(norms)

def rel_dist(x,y):
	x = np.asarray(x)
	y = np.asarray(y)
	return np.linalg.norm(x-y)/np.linalg.norm(np.mean((x,y)))




import unittest
from jitcxde_common.check import CheckEnvironment, checker

class SomeChecks(CheckEnvironment):
	def __init__(self,fail=True):
		self.invoked = []
		self.fail = fail
	
	@checker
	def A(self):
		self.invoked.append(1)
	
	@checker
	def B(self):
		self.invoked.append(2)
		self._check_assert( not self.fail, "Check A failed." )
	
	@checker
	def C(self):
		self.invoked.append(3)
	
	@checker
	def D(self):
		self.invoked.append(4)
		self._check_assert( not self.fail, "Check D failed." )

	def E(self):
		raise AssertionError("This method should not be run")

class DifferentChecks(SomeChecks):
	def A(self):
		raise AssertionError("This method should not be run")
	
	@checker
	def B(self):
		self.invoked.append(5)
	
	@checker
	def C(self):
		self.invoked.append(6)
		self._check_assert( not self.fail, "Check C failed." )

class TestChecks(unittest.TestCase):
	def test_default(self):
		X = SomeChecks(fail=True)
		with self.assertRaises(ValueError):
			X.check()
		assert (2 in X.invoked) ^ (4 in X.invoked)
	
	def test_fail_slow(self):
		X = SomeChecks(fail=True)
		with self.assertRaises(ValueError):
			X.check(fail_fast=False)
		self.assertListEqual( sorted(X.invoked), [1,2,3,4] )
	
	def test_success(self):
		X = SomeChecks(fail=False)
		X.check()
		self.assertListEqual( sorted(X.invoked), [1,2,3,4] )

class TestInheritance(unittest.TestCase):
	def test_default(self):
		X = DifferentChecks(fail=True)
		with self.assertRaises(ValueError):
			X.check()
		assert (4 in X.invoked) ^ (6 in X.invoked)
		assert all( i not in X.invoked for i in [1,2,3] )
	
	def test_fail_slow(self):
		X = DifferentChecks(fail=True)
		with self.assertRaises(ValueError):
			X.check(fail_fast=False)
		self.assertListEqual( sorted(X.invoked), [4,5,6] )
	
	def test_success(self):
		X = DifferentChecks(fail=False)
		X.check()
		self.assertListEqual( sorted(X.invoked), [4,5,6] )

if __name__ == "__main__":
	unittest.main(buffer=True)

